@@ -1,19 +1,9 @@
 import django
 import pytest
 from crispy_forms import __version__
-from crispy_forms.bootstrap import Field, InlineCheckboxes
+from crispy_forms.bootstrap import Field, InlineCheckboxes, UneditableField
 from crispy_forms.helper import FormHelper
-from crispy_forms.layout import (
-    HTML,
-    ButtonHolder,
-    Column,
-    Div,
-    Fieldset,
-    Layout,
-    MultiField,
-    Row,
-    Submit,
-)
+from crispy_forms.layout import HTML, Column, Fieldset, Layout, Row, Submit
 from crispy_forms.utils import render_crispy_form
 from django import forms
 from django.forms.models import formset_factory, modelformset_factory
@@ -26,16 +16,18 @@ from django.urls import reverse
 from django.utils.translation import gettext_lazy as _
 
 from .forms import (
+    AdvancedFileForm,
     CheckboxesSampleForm,
     CrispyEmptyChoiceTestModel,
     CrispyTestModel,
+    FileForm,
     SampleForm,
     SampleForm2,
     SampleForm3,
     SampleForm4,
     SampleForm5,
     SampleForm6,
-    SimpleCheckboxSampleForm,
+    SelectSampleForm,
 )
 from .test_settings import TEMPLATE_DIRS
 from .utils import contains_partial, parse_expected, parse_form
@@ -209,7 +201,8 @@ def test_layout_fieldset_row_html_with_unicode_fieldnames():
     assert 'test-fieldset="123"' in html
     assert 'id="row_passwords"' in html
     assert html.count("<label") == 6
-    assert 'class="row rows"' in html
+
+    assert 'class="form-row rows"' in html
     assert "Hello!" in html
     assert "testLink" in html
 
@@ -252,6 +245,60 @@ def test_change_layout_dynamically_delete_field():
     assert "email" not in html
 
 
+def test_column_has_css_classes():
+    template = Template(
+        """
+        {% load crispy_forms_tags %}
+        {% crispy form form_helper %}
+    """
+    )
+
+    form = SampleForm()
+    form_helper = FormHelper()
+    form_helper.add_layout(
+        Layout(
+            Fieldset(
+                "Company Data",
+                "is_company",
+                "email",
+                "password1",
+                "password2",
+                css_id="multifield_info",
+            ),
+            Column("first_name", "last_name"),
+        )
+    )
+
+    c = Context({"form": form, "form_helper": form_helper})
+    html = template.render(c)
+    assert html.count("col-md") == 1
+
+
+def test_bs4_column_css_classes():
+    template = Template(
+        """
+        {% load crispy_forms_tags %}
+        {% crispy form form_helper %}
+    """
+    )
+
+    form = SampleForm()
+    form_helper = FormHelper()
+    form_helper.add_layout(
+        Layout(
+            Column("first_name", "last_name"),
+            Column("first_name", "last_name", css_class="col-sm"),
+            Column("first_name", "last_name", css_class="mb-4"),
+        )
+    )
+
+    c = Context({"form": form, "form_helper": form_helper})
+    html = template.render(c)
+
+    assert html.count("col-md") == 2
+    assert html.count("col-sm") == 1
+
+
 def test_formset_layout():
     SampleFormSet = formset_factory(SampleForm, extra=3)
     formset = SampleFormSet()
@@ -313,6 +360,8 @@ def test_formset_layout():
     assert html.count("Note for first form only") == 1
     assert html.count("row") == 3
 
+    assert html.count("form-group") == 18
+
 
 def test_modelformset_layout():
     CrispyModelFormSet = modelformset_factory(
@@ -423,122 +472,6 @@ def test_choice_with_none_is_selected():
     assert "checked" in html
 
 
-@override_settings(CRISPY_CLASS_CONVERTERS=CONVERTERS)
-def test_layout_composition(settings):
-    form_helper = FormHelper()
-    form_helper.add_layout(
-        Layout(
-            Layout(
-                MultiField(
-                    "Some company data",
-                    "is_company",
-                    "email",
-                    css_id="multifield_info",
-                ),
-            ),
-            Column(
-                "first_name",
-                # 'last_name', Missing a field on purpose
-                css_id="column_name",
-                css_class="columns",
-            ),
-            ButtonHolder(
-                Submit("Save", "Save", css_class="button white"),
-            ),
-            Div(
-                "password1",
-                "password2",
-                css_id="custom-div",
-                css_class="customdivs",
-            ),
-        )
-    )
-
-    template = Template(
-        """
-            {% load crispy_forms_tags %}
-            {% crispy form form_helper %}
-        """
-    )
-    c = Context({"form": SampleForm(), "form_helper": form_helper})
-    html = template.render(c)
-
-    # Bootstrap 4 does not contain a multifield template
-    if django.VERSION >= (5, 0):
-        # Added 'aria-describedby' for fields with help_text
-        # https://docs.djangoproject.com/en/5.0/releases/5.0/#forms
-        expected = parse_expected(
-            "bootstrap3/test_layout/test_layout_composition_gte50.html"
-        )
-    else:
-        expected = parse_expected("bootstrap3/test_layout/test_layout_composition.html")
-    assert parse_html(html) == expected
-
-
-@override_settings(CRISPY_CLASS_CONVERTERS=CONVERTERS)
-def test_second_layout_multifield_column_buttonholder_submit_div(settings):
-    form_helper = FormHelper()
-    form_helper.add_layout(
-        Layout(
-            MultiField(
-                "Some company data",
-                "is_company",
-                "email",
-                css_id="multifield_info",
-                title="multifield_title",
-                multifield_test="123",
-            ),
-            Column(
-                "first_name",
-                "last_name",
-                css_id="column_name",
-                css_class="columns",
-            ),
-            ButtonHolder(
-                Submit(
-                    "Save the world",
-                    "{{ value_var }}",
-                    css_class="button white",
-                    data_id="test",
-                    data_name="test",
-                ),
-                Submit("store", "Store results"),
-            ),
-            Div(
-                "password1",
-                "password2",
-                css_id="custom-div",
-                css_class="customdivs",
-                test_markup="123",
-            ),
-        )
-    )
-
-    template = Template(
-        """
-            {% load crispy_forms_tags %}
-            {% crispy form form_helper %}
-        """
-    )
-    c = Context({"form": SampleForm(), "form_helper": form_helper, "value_var": "Save"})
-    html = template.render(c)
-
-    # Bootstrap 4 does not contain a multifield template
-    if django.VERSION >= (5, 0):
-        # Added 'aria-describedby' for fields with help_text
-        # https://docs.djangoproject.com/en/5.0/releases/5.0/#forms
-        expected = parse_expected(
-            "bootstrap3/test_layout/"
-            "test_second_layout_multifield_column_buttonholder_submit_div_gte50.html"
-        )
-    else:
-        expected = parse_expected(
-            "bootstrap3/test_layout/"
-            "test_second_layout_multifield_column_buttonholder_submit_div.html"
-        )
-    assert parse_html(html) == expected
-
-
 @override_settings(
     TEMPLATES=[
         {
@@ -553,7 +486,7 @@ def test_second_layout_multifield_column_buttonholder_submit_div(settings):
         }
     ]
 )
-def test_keepcontext_context_manager(settings):
+def test_keepcontext_context_manager():
     # Test case for issue #180
     # Apparently it only manifest when using render_to_response this exact way
     form = CheckboxesSampleForm()
@@ -568,13 +501,11 @@ def test_keepcontext_context_manager(settings):
         request=None, template_name="crispy_render_template.html", context=context
     )
 
-    assert response.content.count(b"checkbox-inline") == 3
+    assert response.content.count(b"custom-control-inline") == 3
+    assert response.content.count(b"custom-checkbox") > 0
 
 
-@pytest.mark.skipif(
-    __version__[0] == "1", reason='#1262 fixed required attributes and empty for=""'
-)
-def test_multiple_checkboxes_bs3():
+def test_use_custom_control_is_used_in_checkboxes():
     form = CheckboxesSampleForm()
     form.helper = FormHelper()
     form.helper.layout = Layout(
@@ -582,50 +513,111 @@ def test_multiple_checkboxes_bs3():
         InlineCheckboxes("alphacheckboxes"),
         "numeric_multiple_checkboxes",
     )
+    # form.helper.use_custom_control take default value which is True
     assert parse_form(form) == parse_expected(
-        "bootstrap3/test_layout/test_multiple_checkboxes.html"
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_checkboxes_true.html"
     )
 
+    form.helper.use_custom_control = True
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_checkboxes_true.html"
+    )
 
-def test_no_label_checkboxes_bs3():
-    form = SimpleCheckboxSampleForm()
+    form.helper.use_custom_control = False
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/"
+        "test_use_custom_control_is_used_in_checkboxes_false.html"
+    )
+
+    form = CheckboxesSampleForm({})
     form.helper = FormHelper()
-    # no form-control class when labels are rendered
-    html = render_crispy_form(form)
-    assert html.count("form-control") == 0
-    form.helper.form_show_labels = False
-    # no labels or form-control class when labels are hidden
-    html = render_crispy_form(form)
-    assert html.count("<label ") == 0
-    assert html.count("form-control") == 0
+    form.helper.layout = Layout(
+        "checkboxes",
+        InlineCheckboxes("alphacheckboxes"),
+        "numeric_multiple_checkboxes",
+    )
+    if django.VERSION < (5, 0):
+        expected = (
+            "bootstrap4/test_layout/"
+            "test_use_custom_control_is_used_in_checkboxes_true_failing_lt50.html"
+        )
+    else:
+        expected = (
+            "bootstrap4/test_layout/"
+            "test_use_custom_control_is_used_in_checkboxes_true_failing.html"
+        )
+    assert parse_form(form) == parse_expected(expected)
 
 
-@pytest.mark.skipif(__version__[0] == "1", reason="#1262 fixed required attributes.")
-def test_radio_bs3():
+def test_use_custom_control_is_used_in_radio():
     form = SampleForm5()
     form.helper = FormHelper()
     form.helper.layout = Layout(
         "radio_select",
     )
-    assert parse_form(form) == parse_expected("bootstrap3/test_layout/test_radio.html")
+    # form.helper.use_custom_control take default value which is True
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_true.html"
+    )
+
+    form.helper.use_custom_control = True
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_true.html"
+    )
+
+    form.helper.use_custom_control = False
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_false.html"
+    )
+
+    form = SampleForm5({})
+    form.helper = FormHelper()
+    form.helper.layout = Layout(
+        "radio_select",
+    )
+    if django.VERSION < (5, 0):
+        expected = (
+            "bootstrap4/test_layout/"
+            "test_use_custom_control_is_used_in_radio_true_failing_lt50.html"
+        )
+    else:
+        expected = (
+            "bootstrap4/test_layout/"
+            "test_use_custom_control_is_used_in_radio_true_failing.html"
+        )
+    assert parse_form(form) == parse_expected(expected)
 
 
-def test_form_inline():
+@pytest.mark.parametrize(
+    "use_custom_control, expected_html",
+    [
+        (True, "bootstrap4/test_layout/test_use_custom_control_in_select_true.html"),
+        (False, "bootstrap4/test_layout/test_use_custom_control_in_select_false.html"),
+    ],
+)
+def test_use_custom_control_in_select(use_custom_control, expected_html):
+    form = SelectSampleForm()
+
+    form.helper = FormHelper()
+    form.helper.template_pack = "bootstrap4"
+    form.helper.layout = Layout("select")
+    form.helper.use_custom_control = use_custom_control
+
+    assert parse_form(form) == parse_expected(expected_html)
+
+
+def test_bootstrap4_form_inline():
     form = SampleForm()
     form.helper = FormHelper()
     form.helper.form_class = "form-inline"
-    form.helper.field_template = "bootstrap3/layout/inline_field.html"
-    form.helper.layout = Layout(
-        "email",
-        "password1",
-        "last_name",
-    )
+    form.helper.field_template = "bootstrap4/layout/inline_field.html"
+    form.helper.layout = Layout("email", "password1", "last_name")
 
     html = render_crispy_form(form)
     assert html.count('class="form-inline"') == 1
-    assert html.count('class="form-group"') == 3
+    assert html.count('class="input-group"') == 3
     assert html.count('<label for="id_email" class="sr-only') == 1
-    assert html.count('id="div_id_email" class="form-group"') == 1
+    assert html.count('id="div_id_email" class="input-group"') == 1
     assert html.count('placeholder="email"') == 1
     assert html.count("</label> <input") == 3
 
@@ -648,19 +640,184 @@ def test_update_attributes_class():
     assert html.count(' class="hello hello2') == 1
 
 
-def test_multiple_fields():
+@override_settings(CRISPY_CLASS_CONVERTERS=CONVERTERS)
+def test_file_field():
+    form = FileForm()
+    form.helper = FormHelper()
+    form.helper.field_class = "col-lg-9 mb-2"
+    form.helper.layout = Layout("clearable_file")
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_clearable_custom_control.html"
+    )
+
+    form.helper.use_custom_control = False
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_clearable.html"
+    )
+
+    form.helper.use_custom_control = True
+    form.helper.layout = Layout("file_field")
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_custom_control.html"
+    )
+
+    form.helper.use_custom_control = False
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_default.html"
+    )
+
+
+def test_file_field_with_custom_class():
+    form = AdvancedFileForm()
+    form.helper = FormHelper()
+    form.helper.layout = Layout("clearable_file")
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_with_custom_class_clearable.html"
+    )
+
+    form.helper.layout = Layout("file_field")
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_with_custom_class.html"
+    )
+
+
+def test_form_control_size():
+    "CSS classes form-control and form-control-lg are both required"
+    form = SampleForm()
+    form.helper = FormHelper()
+    form.helper.layout = Layout(Field("first_name", css_class="form-control-lg"))
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_form_control_size.html"
+    )
+
+
+@pytest.mark.skipif(
+    __version__[0] == "1",
+    reason="#1224 changed editable field behaviour post crispy forms 1.x",
+)
+def test_uneditable_field():
+    form = SampleForm()
+    form.helper = FormHelper()
+    form.helper.layout = Layout(
+        UneditableField("first_name"),
+    )
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_uneditable_field.html"
+    )
+
+
+@pytest.mark.parametrize(
+    "use_custom_control, expected_html",
+    [
+        (
+            True,
+            "bootstrap4/test_layout/"
+            "test_use_custom_control_in_uneditable_select_true.html",
+        ),
+        (
+            False,
+            "bootstrap4/test_layout/"
+            "test_use_custom_control_in_uneditable_select_false.html",
+        ),
+    ],
+)
+@pytest.mark.skipif(
+    __version__[0] == "1",
+    reason="#1224 changed editable field behaviour post crispy forms 1.x",
+)
+def test_use_custom_control_in_uneditable_select(use_custom_control, expected_html):
+    form = SelectSampleForm()
+    form.helper = FormHelper()
+    form.helper.template_pack = "bootstrap4"
+    form.helper.layout = Layout(UneditableField("select"))
+    form.helper.use_custom_control = use_custom_control
+    assert parse_form(form) == parse_expected(expected_html)
+
+
+def test_multiple_fields(settings):
     "Field can accept any number of fields and apply the kwargs to all fields"
     form = SampleForm()
     form.helper = FormHelper()
     form.helper.layout = Layout(
         Field("first_name", "last_name", css_class="form-control-lg")
     )
+    template_pack = settings.CRISPY_TEMPLATE_PACK
     assert parse_form(form) == parse_expected(
-        "bootstrap3/test_layout/test_multiple_fields.html"
+        f"{template_pack}/test_layout/test_multiple_fields.html"
+    )
+
+
+def test_fundamentals():
+    """
+    This is the example that is in the `crispy_tag_forms` docs.
+    """
+
+    class ExampleForm(forms.Form):
+        like_website = forms.TypedChoiceField(
+            label="Do you like this website?",
+            choices=((1, "Yes"), (0, "No")),
+            coerce=lambda x: bool(int(x)),
+            widget=forms.RadioSelect,
+            initial="1",
+            required=True,
+        )
+
+        favorite_food = forms.CharField(
+            label="What is your favorite food?",
+            max_length=80,
+            required=True,
+        )
+
+        favorite_color = forms.CharField(
+            label="What is your favorite color?",
+            max_length=80,
+            required=True,
+        )
+
+        favorite_number = forms.IntegerField(
+            label="Favorite number",
+            required=False,
+        )
+
+        notes = forms.CharField(
+            label="Additional notes or feedback",
+            required=False,
+        )
+
+        def __init__(self, *args, **kwargs):
+            super().__init__(*args, **kwargs)
+            self.helper = FormHelper()
+            self.helper.form_id = "id-exampleForm"
+            self.helper.form_class = "blueForms"
+            self.helper.form_method = "post"
+            self.helper.form_action = "submit_survey"
+
+            self.helper.add_input(Submit("submit", "Submit"))
+
+    form = ExampleForm()
+    context = {"csrf_token": "NotARealToken"}
+    html = render_crispy_form(form, context=context)
+    assert parse_html(html) == parse_expected(
+        "bootstrap4/test_layout/test_fundamentals_example.html"
+    )
+
+
+@pytest.mark.skipif(
+    __version__[0] == "1", reason="#1250 removed custom-control post crispy forms 1.x"
+)
+def test_table_inline_formset_checkbox():
+    class TestForm(forms.Form):
+        box_one = forms.CharField(label="box one", widget=forms.CheckboxInput())
+        box_two = forms.CharField(label="box two", widget=forms.CheckboxInput())
+
+    formset = formset_factory(TestForm)
+    formset.helper = FormHelper()
+    formset.helper.template = "bootstrap4/table_inline_formset.html"
+    assert parse_form(formset) == parse_expected(
+        "bootstrap4/test_layout/test_inline_formset_checkbox.html"
     )
 
 
-@pytest.mark.skipif(__version__[0] == "1", reason="#1262 fixed required attributes.")
 def test_radio_attrs():
     class TestForm(forms.Form):
         radios = forms.ChoiceField(
@@ -678,5 +835,5 @@ def test_radio_attrs():
     form = TestForm()
     form.helper = FormHelper()
     assert parse_form(form) == parse_expected(
-        "bootstrap3/test_layout/test_radio_attrs.html"
+        "bootstrap4/test_layout/test_radio_attrs.html"
     )
