@@ -27,11 +27,7 @@ def contains_partial(haystack, needle, ignore_needle_children=False):
     ):
         return True
     return any(
-        contains_partial(
-            child,
-            needle,
-            ignore_needle_children=ignore_needle_children,
-        )
+        contains_partial(child, needle, ignore_needle_children=ignore_needle_children)
         for child in haystack.children
         if isinstance(child, Element)
     )
